from pydantic import validate_arguments
import random

from .config import (
    damage_variant,
    attack_categories_power,
    rank_class_power,
    one_star_power
)

@validate_arguments
def calculate_attack_damage(
    category: int,
    level: int,
    attack_power: int,
    rank_class: int = 0,
    stars: int = 0
) -> dict:
    initial_damage = attack_categories_power[category]

    rank_class_bonus = 0
    stars_bonus = 0

    if rank_class > 0:
        rank_class_bonus = rank_class_power[rank_class]  

    if stars > 0:
        stars_bonus = one_star_power * stars
        
    damage = 1.5 * (1 + rank_class_bonus + stars_bonus) * (initial_damage * (level ** 1.5 + 10) / 250) + attack_power

    minimum = round(damage_variant[0] * damage)
    maximum = round(damage_variant[1] * damage)
    average = round((minimum + maximum) / 2)

    random_damage = round(
        random.randint(
            int(damage_variant[0] * 100),
            int(damage_variant[1] * 100)
        ) / 100
            *
        damage
    )

    return dict(
        minimum = minimum,
        maximum = maximum,
        average = average,
        random = random_damage
    )

__all__ = [ "calculate_attack_damage" ]


from pydantic import validate_arguments
from pyfilter.tuple_list import FromTupleList
from rich import print
import random

from .dragon import calculate_attack_damage, calculate_status
from .elements import calculate_strongs, calculate_weaknesses

class SimulationTypes:
    NORMAL = 0

def select_best_attack(
    attacks: list[tuple[str, int]],
    enemy_weaknesses: list[str],
    enemy_main_element: str
) -> tuple[str, int]:
    selected_attack: tuple[str, int]
    attacks_with_strong = []
    enemy_strongs = calculate_strongs([ enemy_main_element ])

    for attack in attacks:
        attack_element = attack[0]

        for enemy_weakness in enemy_weaknesses:
            weakness_element = enemy_weakness[0]

            if weakness_element == attack_element:
                attacks_with_strong.append(attack)

    if len(attacks_with_strong) > 0:
        attacks_with_strongs_power_max = max([ attack[1] for attack in attacks_with_strong ])
        selected_attack = FromTupleList(attacks_with_strong).get_with_value(attacks_with_strongs_power_max)

    else:
        attacks_without_null = [
            attack for attack in attacks
            if attack[0] != enemy_main_element
        ]

        attacks_without_weaknesses = [
            attack for attack in attacks_without_null
            if attack[0] in enemy_strongs
        ]

        if len(attacks_without_weaknesses) > 0:
            attacks_without_weaknesses_power_max = max([ attack[1] for attack in attacks_without_weaknesses ])
            attack_without_weaknesses = FromTupleList(attacks_without_weaknesses).get_with_value(attacks_without_weaknesses_power_max)
            selected_attack = attack_without_weaknesses

        else:
            if len(attacks_without_null) > 0:
                attack_power_max = max([ attack[1] for attack in attacks_without_null ])
                attack = FromTupleList(attacks).get_with_value(attack_power_max)
                selected_attack = attack

            else:
                selected_attack = attacks[0]

    return selected_attack

class Entity:
    def __init__(
        self,
        category: int,
        level: int,
        rank_class: int,
        starts: int,
        hp: int,
        elements: list[str],
        attacks: list[tuple[str, int]]
    ) -> None:
        self.__category = category
        self.__level = level
        self.__rank_class = rank_class
        self.__stars = starts

        attack_elements = [ attack[0] for attack in attacks ]

        self.elements = elements
        self.main_element = elements[0]
        self.weaknesses = calculate_weaknesses(self.main_element)
        self.strongs = calculate_strongs(attack_elements)
        self.attacks = attacks
        self.current_hp = int(hp)

    @validate_arguments
    def attack(self, entity, attack: tuple[str, int]):
        attack_element, attack_power = attack

        damage = calculate_attack_damage(
            self.__category,
            self.__level,
            attack_power,
            self.__rank_class,
            self.__stars
        )["random"]

        is_strong_damage = False
        is_weak_damage = False
        
        attack_strongs = calculate_strongs([ attack_element ])
        attack_weaknesses = calculate_weaknesses(attack_element)
        dragon_main_element_strongs = calculate_strongs([ entity.main_element ])

        damage_type = "normal"

        for strong in attack_strongs:
            if strong in entity.weaknesses:
                is_strong_damage = True
                break

        for weakness in attack_weaknesses:
            if weakness in dragon_main_element_strongs:
                is_weak_damage = True
                break

        if is_strong_damage:
            damage += damage
            damage_type = "strong"
        
        elif is_weak_damage:
            damage -= int(damage / 2)
            damage_type = "weak"

        entity.take_damage(damage)

        return dict(
            damage = damage,
            damage_type = damage_type
        )

    @validate_arguments
    def take_damage(self, damage: int):
        self.current_hp -= damage

        if self.current_hp < 0:
            self.current_hp = 0

    @property
    def is_alive(self) -> bool:
        return self.current_hp > 0

@validate_arguments
def process_entity(
    category: int,
    rarity: str,
    level: int,
    rank_class: int,
    stars: int,
    hp_runes: int,
    damage_runes: int,
    with_tower_bonus: bool,
    extra_hp_multiplier: float,
    extra_damage_multiplier: float,
    elements: list[str],
    attacks: list[tuple[str, int]],
) -> Entity:
    status = calculate_status(
        category,
        rarity,
        level,
        rank_class,
        stars,
        hp_runes,
        damage_runes,
        with_tower_bonus,
        extra_hp_multiplier,
        extra_damage_multiplier,
    )

    hp = status["result"]["hp"]

    entity = Entity(
        category,
        level,
        rank_class,
        stars,
        hp,
        elements,
        attacks
    )

    return entity

class BattleSimulator:
    @validate_arguments
    def __init__(
        self,
        team1_data: list[dict],
        team2_data: list[dict]
    ) -> None:
        team1 = [ process_entity(**entity_data) for entity_data in team1_data ]
        team2 = [ process_entity(**entity_data) for entity_data in team2_data ]

        self.__teams = [team1, team2]
        self.__team_that_starts_index = random.randint(0, 1)

    @validate_arguments
    def simulate_battle(
        self,
        simulation_type: int = SimulationTypes.NORMAL
    ) -> dict:
        current_entity1_index = 0
        current_entity2_index = 0

        start_team_index = self.__team_that_starts_index
        other_team_index = (start_team_index + 1) % 2

        current_entity1 = self.__teams[start_team_index][current_entity1_index]
        current_entity2 = self.__teams[other_team_index][current_entity2_index]

        shifts = []

        if simulation_type == SimulationTypes.NORMAL:
            while True:
                while current_entity1.is_alive:
                    entity1_selected_attack = select_best_attack(
                        current_entity1.attacks,
                        current_entity2.weaknesses,
                        current_entity2.main_element
                    )

                    current_entity1_attack_data = current_entity1.attack(current_entity2, entity1_selected_attack)

                    shifts.append({
                        "team_index": start_team_index,
                        "entity_index": current_entity1_index,
                        "selected_attack": entity1_selected_attack,
                        "damage": current_entity1_attack_data["damage"],
                        "damage_type": current_entity1_attack_data["damage_type"]
                    })

                    if not current_entity2.is_alive:
                        current_entity2_index += 1

                        if current_entity2_index >= len(self.__teams[other_team_index]):
                            return dict(
                                winner_index = start_team_index,
                                shifts = shifts,
                                living_dragons = [
                                    self.__teams[start_team_index].index(entity)
                                    for entity in self.__teams[start_team_index][:current_entity1_index] 
                                ]
                            )

                        current_entity2 = self.__teams[other_team_index][current_entity2_index]
                        
                    entity2_selected_attack = select_best_attack(
                        current_entity2.attacks,
                        current_entity1.weaknesses,
                        current_entity1.main_element
                    )

                    current_entity2_attack_data = current_entity2.attack(current_entity1, entity2_selected_attack)

                    shifts.append({
                        "team_index": other_team_index,
                        "entity_index": current_entity2_index,
                        "selected_attack": entity2_selected_attack,
                        "damage": current_entity2_attack_data["damage"],
                        "damage_type": current_entity2_attack_data["damage_type"]
                    })

                else:
                    current_entity1_index += 1

                    if current_entity1_index >= len(self.__teams[start_team_index]):
                        return dict(
                            winner_index = other_team_index,
                            shifts = shifts,
                            living_dragons =  [
                                self.__teams[other_team_index].index(entity)
                                for entity in self.__teams[other_team_index][:current_entity2_index] 
                            ]
                        )

                    current_entity1 = self.__teams[start_team_index][current_entity1_index]



DRAGON_ELEMENTS_TYPES = {
    "ph": "pysical", "w": "sea", "p": "nature","f": "flame","d": "dark","e": "terra",
    "el": "electric", "m": "metal","i": "ice","wr": "war",
    "l": "legend", "li": "light", "pu": "pure","bt": "beauty",
    "ch": "chaos", "mg": "magic", "hp": "happy","dr": "dream",
    "so": "soul", "pr": "primal", "wd": "wind","ti": "time"
}

__all__ = [
    "DRAGON_ELEMENTS_TYPES"
]


from setuptools import setup, find_packages

with open("README.md", "r") as file:
    readme_content = file.read()

# with open(
#     "requirements.txt",
#     mode="r",
#     encoding="UTF-16"
# ) as file:
#     requirements = [ line.split("==")[0].strip() for line in file.readlines() ]

requirements = ['anyio', 'bleach', 'certifi', 'charset-normalizer', 'docutils', 'h11', 'httpcore', 'httpx', 'idna', 'importlib-metadata', 'jaraco.classes', 'keyring', 'markdown-it-py', 'mdurl', 'more-itertools', 'pkginfo', 'pydantic', 'Pygments', 'python-filter', 'pywin32-ctypes', 'readme-renderer', 'requests', 'requests-toolbelt', 'rfc3986', 'rich', 'six', 'sniffio', 'twine', 'typing_extensions', 'urllib3', 'webencodings', 'zipp']

setup(
    name="dragon-city-utils",
    version="1.23",
    license="MIT License",
    author="Marcuth",
    long_description=readme_content,
    long_description_content_type="text/markdown",
    author_email="marcuth2006@gmail.com",
    keywords="dragoncity dcutils tools",
    description=f"Utilities and tools for things related to Dragon City",
    packages=["dcutils"] + [ "dcutils/" + x for x in find_packages("dcutils") ],
    install_requires=requirements
)


from pydantic import validate_arguments
import re

class DragonUrlParser:
    @classmethod
    @validate_arguments
    def get_image_name(cls, url: str) -> str | None:
        match = re.search(r"\/(basic_|thumb_|ui_)?(\d+)_([\w_]+)", url)

        if match:
            id = match.group(2)
            image_name_without_id = match.group(3)
            raw_image_name = f"{id}_{image_name_without_id}"
            void_str = ""

            image_name_with_skin = re.sub(
                r"_\d+",
                void_str,
                raw_image_name
            )

            image_name_without_skin = re.sub(
                r"_skin\d+",
                void_str,
                image_name_with_skin
            )

            return image_name_without_skin

    @classmethod
    @validate_arguments
    def get_id(cls, url: str) -> int | None:
        match = re.search(r"\/(basic_|thumb_|ui_)?(\d+)_", url)

        if match:
            id = int(match.group(2))
            return id
    
    @classmethod
    @validate_arguments
    def get_phase(cls, url: str) -> int | None:
        match = re.search(r"(\d+)@2x\.(png|swf)|(\d+)\.(png|swf)|(\d+)_HD_tweened_dxt5.zip|(\d+)_HD_spine-3-8-59_dxt5.zip", url)

        if match:
            phase = int(
                match.group(1) or
                match.group(3) or
                match.group(5) or
                match.group(6))

            return phase

    @classmethod
    @validate_arguments
    def get_skin(cls, url: str) -> str | None:
        match = re.search(r"_skin\d+", url)

        if match:
            skin = match.group(0)[1:]
            return skin

    @classmethod
    @validate_arguments
    def get_image_qualitity(cls, url: str) -> str:
        match = re.search(r"@\d+x", url)

        if match:
            image_qualitity = match.group(0)
            return image_qualitity

        return ""

    @classmethod
    @validate_arguments
    def from_sprite(cls, url: str) -> dict:
        return dict(
            id = cls.get_id(url),
            image_name = cls.get_image_name(url),
            phase = cls.get_phase(url),
            skin = cls.get_skin(url),
            image_qualitity = cls.get_image_qualitity(url)
        )

    @classmethod
    @validate_arguments
    def from_thumb(cls, url: str) -> dict:
        return dict(
            id = cls.get_id(url),
            image_name = cls.get_image_name(url),
            phase = cls.get_phase(url),
            skin = cls.get_skin(url)
        )

    @classmethod
    @validate_arguments
    def from_flash_animation(cls, url: str) -> dict:
        return dict(
            id = cls.get_id(url),
            image_name = cls.get_image_name(url),
            phase = cls.get_phase(url),
            skin = cls.get_skin(url)
        )

    @classmethod
    @validate_arguments
    def from_spine_animation(cls, url: str) -> dict:
        return dict(
            id = cls.get_id(url),
            image_name = cls.get_image_name(url),
            phase = cls.get_phase(url),
            skin = cls.get_skin(url)
        )

__all__ = [ "DragonUrlParser" ]


from pydantic import validate_arguments

from .config import (
    category_9_dragons_hp_fix,
    dragon_rarity_power,
    category_and_level_power,
    damage_rune_power,
    hp_rune_power,
    rank_class_power,
    hp_tower_power,
    damage_tower_power
)

@validate_arguments
def calculate_status(
    category: int,
    rarity: str,
    level: int = 1,
    rank_class: int = 0,
    stars: int = 0,
    hp_runes: int = 0,
    damage_runes: int = 0,
    with_tower_bonus: bool = False,
    extra_hp_multiplier: float = 0.0,
    extra_damage_multiplier: float = 0.0
) -> dict:
    initial_status = category_and_level_power[category][level]
    initial_hp = initial_status["hp"]
    initial_damage = initial_status["damage"]

    hp = float(initial_hp)
    damage = float(initial_damage)

    rank_class_hp_bonus = 0
    rank_class_damage_bonus = 0
    stars_hp_bonus = 0
    stars_damage_bonus = 0
    runes_hp_bonus = 0
    runes_damage_bonus = 0
    tower_hp_bonus = 0
    tower_damage_bonus = 0
    extra_hp_bonus = 0
    extra_damage_bonus = 0

    if rank_class > 0:
        rank_class_factor = rank_class_power[rank_class] / 100
        rank_class_hp_bonus = initial_hp * rank_class_factor
        rank_class_damage_bonus = initial_damage * rank_class_factor

    if stars > 0:
        stars_factor = dragon_rarity_power[rarity][stars] / 100
        stars_hp_bonus = initial_hp * stars_factor
        stars_damage_bonus = initial_damage * stars_factor

    if hp_runes > 0:
        runes_hp_bonus = initial_hp * (hp_rune_power * hp_runes)
    
    if damage_runes > 0:
        runes_damage_bonus = initial_damage * (damage_rune_power * damage_runes)

    if with_tower_bonus:
        tower_hp_bonus = (hp + (stars_hp_bonus + runes_hp_bonus + rank_class_hp_bonus)) * hp_tower_power
        tower_damage_bonus = (damage + (rank_class_damage_bonus + stars_damage_bonus + runes_damage_bonus)) * damage_tower_power

    if extra_hp_multiplier > 0.0:
        extra_hp_bonus = (hp + (stars_hp_bonus + runes_hp_bonus + rank_class_hp_bonus)) * extra_hp_multiplier

    if extra_damage_multiplier > 0.0:
        extra_damage_bonus = (damage + (stars_damage_bonus + runes_damage_bonus + rank_class_damage_bonus)) * extra_damage_multiplier

    hp += stars_hp_bonus + runes_hp_bonus + rank_class_hp_bonus + tower_hp_bonus + extra_hp_bonus
    damage += rank_class_damage_bonus + stars_damage_bonus + runes_damage_bonus + tower_damage_bonus + extra_damage_bonus

    if category == 9:
        hp += hp * category_9_dragons_hp_fix

    hp = round(hp)
    damage = round(damage)

    return dict(
        result = dict(
            hp = hp,
            damage = damage
        ),
        initial = dict(
            hp = round(initial_hp if category != 9 else initial_hp * (category_9_dragons_hp_fix + 1)),
            damage = round(initial_damage), 
        ),
        bonus = dict(
            rank_class = dict(
                hp = round(rank_class_hp_bonus),
                damage = round(rank_class_damage_bonus)
            ),
            stars = dict(
                hp = round(stars_hp_bonus),
                damage = round(stars_damage_bonus)
            ),
            runes = dict(
                hp = round(runes_hp_bonus),
                damage = round(runes_damage_bonus)
            ),
            tower = dict(
                hp = round(tower_hp_bonus),
                damage = round(tower_damage_bonus)
            ),
            extra = dict(
                hp = round(extra_hp_bonus),
                damage = round(extra_damage_bonus)
            )
        )
    )

__all__ = [ "calculate_status" ]